/**Copyright or (C) or Copr. Institut Mines Telecom, Telecom ParisTech, Ludovic Apvrille

ludovic.apvrille AT telecom-paristech.fr

This software is a computer program whose purpose is to allow malware analysis

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

/**
 * Class ClusterAlgoMultithreadManager
 * Managing the multithread execution of algorithms of cluster
 * Creation: 17/07/2013
 * @version 1.0 17/07/2013
 * @author Ludovic APVRILLE
 * @see
 */

package cluster;

import myutil.*;


public class ClusterAlgoMultithreadManager  {
	
	public static final int MULTITHREADED_THRESHOLD = 25;
	public final static int DEFAULT_NB_OF_THREADS = 4; // Must be at least 1
	
	private Cluster cluster;
	
	private ClusterAlgoMultithread[] threads;
	private int nbOfThreads;
	
	public ClusterAlgoMultithreadManager(Cluster _cluster) {
		cluster = _cluster;
		
		int size = cluster.getElements().size();
		
		if (size < MULTITHREADED_THRESHOLD) {
			nbOfThreads = 1;
		} else {
			try {
				nbOfThreads = Runtime.getRuntime().availableProcessors();
			} catch (Exception e) {
				nbOfThreads = DEFAULT_NB_OF_THREADS;
			}
		}
		
	}
	
	public void createThreads(){
		if (threads == null) {
			threads = new ClusterAlgoMultithread[nbOfThreads];
			
			for (int i=0; i<nbOfThreads; i++) {
				ClusterAlgoMultithread t = new ClusterAlgoMultithread(i, cluster);
				threads[i] = t;
				t.start();
			}
		}
	}
	
	
	public void computeClosestElements(int nbOfElements, Cluster[] withClusters) {
		int i;
		
		createThreads();
	
		int size = cluster.getElements().size();
		
		TraceManager.addDev("Creating " + nbOfThreads + " threads to compute " + size + " elements");
		
		// Launching the threads
		for (i=0; i<nbOfThreads; i++) {
			threads[i].algo = ClusterAlgoMultithread.ALGO_CLOSEST;
			threads[i].withClusters = withClusters;
			threads[i].nbOfElements = nbOfElements;
			threads[i].updatePercentage = (i == 0);
			threads[i].startIndex = (size/nbOfThreads) * (i);
			threads[i].endIndex = ((size/nbOfThreads) * (i+1));
			if (i == (nbOfThreads - 1)) {
				threads[i].endIndex = size;
			}
			threads[i].executeAlgo();
			
		}
		
		waitForThreadsTermination();
	}
	
	public void computeClosestElementsWithProperties(int nbOfElements, int nbOfProperties, Cluster[] withClusters) {
		int i;
		
		createThreads();
	
		int size = cluster.getElements().size();
		
		TraceManager.addDev("Creating " + nbOfThreads + " threads to compute " + size + " elements");
		
		// Launching the threads
		for (i=0; i<nbOfThreads; i++) {
			threads[i].algo = ClusterAlgoMultithread.ALGO_CLOSEST_PROPERTIES;
			threads[i].withClusters = withClusters;
			threads[i].nbOfElements = nbOfElements;
			threads[i].nbOfProperties = nbOfProperties;
			//threads[i].nbOfElements = nbOfProperties;
			threads[i].updatePercentage = (i == 0);
			threads[i].startIndex = (size/nbOfThreads) * (i);
			threads[i].endIndex = ((size/nbOfThreads) * (i+1));
			if (i == (nbOfThreads - 1)) {
				threads[i].endIndex = size;
			}
			threads[i].executeAlgo();
			
		}
		
		waitForThreadsTermination();
		
	}

    public void computeClosestElementsWithProperties_FirstVersion(int nbOfElements, int nbOfProperties, Cluster[] withClusters) {
		int i;
		
		createThreads();
	
		int size = cluster.getElements().size();
		
		TraceManager.addDev("Creating " + nbOfThreads + " threads to compute " + size + " elements");
		
		// Launching the threads
		for (i=0; i<nbOfThreads; i++) {
			threads[i].algo = ClusterAlgoMultithread.ALGO_CLOSEST_PROPERTIES;
			threads[i].withClusters = withClusters;
			threads[i].nbOfElements = nbOfElements;
			//threads[i].nbOfProperties = nbOfProperties;
			threads[i].nbOfElements = nbOfProperties;
			threads[i].updatePercentage = (i == 0);
			threads[i].startIndex = (size/nbOfThreads) * (i);
			threads[i].endIndex = ((size/nbOfThreads) * (i+1));
			if (i == (nbOfThreads - 1)) {
				threads[i].endIndex = size;
			}
			threads[i].executeAlgo();
			
		}
		
		waitForThreadsTermination();
		
	}
	
	public void computeCorrelations(int nbOfProperties) {
		int i;
		
		createThreads();
	
		int size = nbOfProperties;
		
		TraceManager.addDev("Creating " + nbOfThreads + " threads to compute " + nbOfProperties + " properties");
		
		// Launching the threads
		for (i=0; i<nbOfThreads; i++) {
			threads[i].algo = ClusterAlgoMultithread.ALGO_CORRELATIONS;
			threads[i].updatePercentage = (i == 0);
			threads[i].startIndex = (size/nbOfThreads) * (i);
			threads[i].endIndex = ((size/nbOfThreads) * (i+1));
			if (i == (nbOfThreads - 1)) {
				threads[i].endIndex = size;
			}
			threads[i].executeAlgo();
			
		}
		
		waitForThreadsTermination();
		
	}
	
	public void waitForThreadsTermination() {
		// Waiting for their termination
		for (int i=0; i<nbOfThreads; i++) {
			threads[i].waitForAlgoDone();
			TraceManager.addDev("Thread #" + i + " completed\n");
		}
		
		TraceManager.addDev("All threads terminated");
	}
	
	public void stopAll() {
		for (int i=0; i<nbOfThreads; i++) {
			threads[i].stopMe();
			threads[i].executeAlgo();
			threads[i] = null;
		}
	}
	
	
	
	
	


}
